#include "spawn.h"

#include "..\algorithms.h"

#include "..\dllcore.cpp"
#include "..\datatypes.cpp"
#include "..\algorithms.cpp"
#include "..\player.cpp"
#include "..\commtypes.cpp"

#define UNASSIGNED 0xffff


//////// Bot list ////////

_linkedlist <botInfo> botlist;

botInfo *findBot(CALL_HANDLE handle)
{
	_listnode <botInfo> *parse = botlist.head;

	while (parse)
	{
		botInfo *item = parse->item;

		if (item->validate(handle))
			return item;

		parse = parse->next;
	}

	return 0;
}


//////// DLL "import" ////////

void botInfo::tell(BotEvent &event)
{
	if (callback && handle)
	{
		event.handle = handle;
		callback(event);
	}
}

Uint32 botInfo::strFirstInstanceOf(char *msg, char *msg2)
{
	Uint32 len1 = STRLEN(msg), len2 = STRLEN(msg2);

	if (len1 < len2)
		return -1;

	for (Uint32 i = 0; i <= (len1 - len2); i++)
	{
		if (msg[i] == msg2[0])
		{
			for (Uint32 i2 = 0; i2 <= len2; i2++)
			{
				if (msg2[i2] == 0)
					return i;

				if (msg[i + i2] != msg2[i2])
					break;
			}
		}
	}

	return -1;
}

String botInfo::strReplace(char *msg, char *from, char *to)
{
	String ret;

	Uint32 fromlen = STRLEN(from);

	Uint32 pos;

	while (true)
	{
		pos = strFirstInstanceOf(msg, from);
		if (pos == -1)
		{
			/// Append the tail and go ///

			ret += msg;
			break;
		}

		ret.append(msg, pos);
		ret += to;
		msg += pos + fromlen;
	}

	return ret;
}

void botInfo::SendString(Player *p,String s)
{
	if(!s.IsEmpty())
	{
		Message = strReplace(s.msg, "%name%", p->name);
		Message = strReplace(Message.msg, "%points%", getString(p->score.killPoints + p->score.flagPoints, 10, 0, false).msg);
		Message = strReplace(Message.msg, "%squad%", p->squad);
		Message = strReplace(Message.msg, "%freq%", getString(p->team, 10, 0, false).msg);
		sendPrivate(p,Message);
	}
}

void botInfo::KillIgnores(Player *p)
{
	il=IgnoreList.head;
	while(il)
	{
		if(il->item->p==p)
		{
			for(al=AreaList.head;al;al=al->next)
			{
				if(al->item->AreaNum == il->item->AreaNum)
				{
					for(i=0;i<5;i++)
					{
						SendString(p,al->item->Leave[i]);
					}
					break;
				}
			}

			if (il->next)//delete the ignore
			{
				il=il->next;
				IgnoreList.kill(il->last);
				continue;
			}
			else
			{
				IgnoreList.kill(il);
				break;
			}
		}
		il=il->next;
	}
}

void botInfo::gotEvent(BotEvent &event)
{
	if (CONNECTION_DENIED) return;

	switch (event.code)
	{
//////// Periodic ////////
	case EVENT_Tick:
		{
		}
		break;
//////// Arena ////////
	case EVENT_ArenaEnter:
		{
			arena = (char*)event.p[0];
			me = (Player*)event.p[1];	// if(me) {/*we are in the arena*/}
			bool biller_online = *(bool*)&event.p[2];

			IgnoreList.clear();
		}
		break;
	case EVENT_ArenaSettings:
		{
			settings = (arenaSettings*)event.p[0];
		}
		break;
	case EVENT_ArenaLeave:
		{
			me = 0;
			killTags();
			IgnoreList.clear();
		}
		break;
	case EVENT_ArenaListEntry:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
	case EVENT_ArenaListEnd:
		{
			char *name = (char*)event.p[0];
			bool current = *(bool*)&event.p[1];
			int population = *(int*)&event.p[2];
		}
		break;
//////// Flag ////////
	case EVENT_FlagGrab:
		{
			Player *p = (Player*)event.p[0];
			Flag *f = (Flag*)event.p[1];
		}
		break;
	case EVENT_FlagDrop:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_FlagMove:
		{
			Flag *f = (Flag*)event.p[0];
		}
		break;
	case EVENT_FlagVictory:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_FlagGameReset:
		{
		}
		break;
	case EVENT_FlagReward:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
//////// Timed ////////
	case EVENT_TimedGameOver:
		{
			Player *p[5];
			for (int i = 0; i < 5; i++)
				p[i] = (Player*)event.p[i];
		}
		break;
//////// Soccer ////////
	case EVENT_SoccerGoal:
		{
			int team = *(int*)&event.p[0];
			int reward = *(int*)&event.p[1];
		}
		break;
	case EVENT_BallMove:
		{
			PBall *ball = (PBall*)event.p[0];
		}
		break;
//////// File ////////
	case EVENT_File:
		{
			char *name = (char*)event.p[0];
		}
		break;
//////// Player ////////
	case EVENT_PlayerEntering:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerMove:
		{
			Player *p = (Player*)event.p[0];

			il=IgnoreList.head;
			while(il)//parse through ignore list
			{
				if(il->item->p==p)//if find a item that matches the player
				{
					for(al=AreaList.head;al;al=al->next)//find the area that matches
					{
						if(al->item->AreaNum == il->item->AreaNum)
						{
							if(!(p->pos.x >= al->item->x1 //check if player is still in area
								&& p->pos.y >= al->item->y1
								&& p->pos.x <= al->item->x2
								&& p->pos.y	<= al->item->y2))
							{
								for(i=0;i<5;i++)
								{
									SendString(p,al->item->Leave[i]);
								}
								al=NULL;//player isnt in area anymore
							}
							break;
						}
					}

					if(al==NULL)//player left the area
					{
						if (il->next)//delete the ignore
						{
							il=il->next;
							IgnoreList.kill(il->last);
							continue;
						}
						else
						{
							IgnoreList.kill(il);
							break;
						}
					}
				}
				il=il->next;
			}

			for(al=AreaList.head;al;al=al->next)//go through area list to see if player is in an area
			{
				if(p->pos.x >= al->item->x1 
					&& p->pos.y >= al->item->y1
					&& p->pos.x <= al->item->x2
					&& p->pos.y <= al->item->y2)
				{
					if ((al->item->freq == -1) || (al->item->freq == -2 && p->team < 100) || (al->item->freq == -3 && p->team > 99) || (al->item->freq==p->team))
					{
						for(il=IgnoreList.head;il;il=il->next)
						{
							if(il->item->p==p && al->item->AreaNum==il->item->AreaNum)
							{break;}
						}

						if(il==NULL)//player enter area
						{
							for(i=0;i<5;i++)
							{
								SendString(p,al->item->Enter[i]);
							}
							IgnoreList.append(new Ignore(p,al->item->AreaNum));//add an ignore
						}
					}
				}
			}
		}
		break;
	case EVENT_PlayerWeapon:
		{
			Player *p = (Player*)event.p[0];
			weaponInfo wi;
			wi.n = *(Uint16*)&event.p[1];
		}
		break;
	case EVENT_WatchDamage:
		{
			Player *p = (Player*)event.p[0];	// player
			Player *k = (Player*)event.p[1];	// attacker
			weaponInfo wi;
			wi.n = *(Uint16*)&event.p[2];
			Uint16 energy = *(Uint16*)&event.p[3];
			Uint16 damage = *(Uint16*)&event.p[4];
		}
		break;
	case EVENT_PlayerDeath:
		{
			Player *p = (Player*)event.p[0],
				   *k = (Player*)event.p[1];
			Uint16 bounty = (Uint16)event.p[2];
		}
		break;
	case EVENT_PlayerScore:
		{
			Player *p = (Player*)event.p[0];
		}
		break;
	case EVENT_PlayerPrize:
		{
			Player *p = (Player*)event.p[0];
			int prize = *(int *)&event.p[1];
		}
		break;
	case EVENT_PlayerShip:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldship = (Uint16)event.p[1];
			Uint16 oldteam = (Uint16)event.p[2];
		}
		break;
	case EVENT_PlayerSpec:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];
			Uint16 oldship = (Uint16)event.p[2];

			KillIgnores(p);
		}
		break;
	case EVENT_PlayerTeam:
		{
			Player *p = (Player*)event.p[0];
			Uint16 oldteam = (Uint16)event.p[1];
			Uint16 oldship = (Uint16)event.p[2];

		}
		break;
	case EVENT_PlayerLeaving:
		{
			Player *p = (Player*)event.p[0];

			killTags(p);
			KillIgnores(p);
		}
		break;
//////// Selfish ////////
	case EVENT_SelfShipReset:
		{
		}
		break;
	case EVENT_SelfPrize:
		{
			Uint16 prize = (Uint16)event.p[0];
			Uint16 count = (Uint16)event.p[1];
		}
		break;
	case EVENT_SelfUFO:
		{
		}
		break;
	case EVENT_PositionHook:
		{
		}
		break;
//////// Bricks ////////
	case EVENT_BrickDropped:
		{
			int x1 = *(int*)&event.p[0];
			int y1 = *(int*)&event.p[1];
			int x2 = *(int*)&event.p[2];
			int y2 = *(int*)&event.p[3];
			int team = *(int*)&event.p[4];
		}
		break;
//////// LVZ ////////
	case EVENT_ObjectToggled:
		{
			objectInfo obj;
			obj.n = (Uint16)event.p[0];
		}
		break;
//////// Turret ////////
	case EVENT_CreateTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
	case EVENT_DeleteTurret:
		{
			Player *turreter = (Player*)event.p[0];
			Player *turretee = (Player*)event.p[1];
		}
		break;
//////// Chat ////////
	case EVENT_Chat:
		{
			int type  = *(int*)&event.p[0];
			int sound = *(int*)&event.p[1];
			Player *p = (Player*)event.p[2];
			char *msg = (char*)event.p[3];

			switch (type)
			{
			case MSG_Arena:
				{
				}
				break;
			case MSG_PublicMacro:		if (!p) break;
				{
				}
				break;
			case MSG_Public:			if (!p) break;
				{
				}
				break;
			case MSG_Team:				if (!p) break;
				{
				}
				break;
			case MSG_TeamPrivate:		if (!p) break;
				{
				}
				break;
			case MSG_Private:			if (!p) break;
				{
				}
				break;
			case MSG_PlayerWarning:		if (!p) break;
				{
				}
				break;
			case MSG_RemotePrivate:
				{
				}
				break;
			case MSG_ServerError:
				{
				}
				break;
			case MSG_Channel:
				{
				}
				break;
			};
		}
		break;
	case EVENT_LocalCommand:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotCommand(p, c);
		}
		break;
	case EVENT_LocalHelp:
		{
			Player *p = (Player*)event.p[0];
			Command *c = (Command*)event.p[1];

			gotHelp(p, c);
		}
		break;
	case EVENT_RemoteCommand:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemote(p, c, o);
		}
		break;
	case EVENT_RemoteHelp:
		{
			char *p = (char*)event.p[0];
			Command *c = (Command*)event.p[1];
			Operator_Level o = *(Operator_Level*)&event.p[2];

			gotRemoteHelp(p, c, o);
		}
		break;
//////// Containment ////////
	case EVENT_Init:
		{
			int major	= HIWORD(*(int*)&event.p[0]);
			int minor	= LOWORD(*(int*)&event.p[0]);
			callback	= (CALL_COMMAND)event.p[1];
			playerlist	= (CALL_PLIST)event.p[2];
			flaglist	= (CALL_FLIST)event.p[3];
			map			= (CALL_MAP)event.p[4];
			bricklist	= (CALL_BLIST)event.p[5];

			if (major > CORE_MAJOR_VERSION)
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin is out of date."));

				CONNECTION_DENIED = true;

				return;
			}
			else if ((major < CORE_MAJOR_VERSION) || (minor < CORE_MINOR_VERSION))
			{
				tell(makeEcho("DLL plugin cannot connect.  This plugin requires the latest version of MERVBot."));

				CONNECTION_DENIED = true;

				return;
			}
			else
			{
				tell(makeEcho("DLL plugin connected."));

				char buffer[250]; //holds the input from the file
				char Path[512]; //file path
				GetCurrentDirectory(512, Path); //grab the path
				strcat(Path, "\\messagebot.ini"); //append the file name

				Area *newArea;
				String m1,m2;
				for(i=1;i<=100;i++)
				{	
					m1="Message";
					m1+=i;

					ZeroMemory(buffer, 10); 
					GetPrivateProfileString(m1.msg, "X1", NULL, buffer, 5, Path);
					if(isNumeric(buffer))
					{
						newArea=new Area;
						newArea->x1=getInteger(buffer,10)*16;
						
						ZeroMemory(buffer, 10); 
						GetPrivateProfileString(m1.msg, "Y1", NULL, buffer, 5, Path);
						if(isNumeric(buffer))
						{
							newArea->y1=getInteger(buffer,10)*16;
							
							ZeroMemory(buffer, 10); 
							GetPrivateProfileString(m1.msg, "X2", NULL, buffer, 5, Path);
							if(isNumeric(buffer))
							{
								newArea->x2=getInteger(buffer,10)*16;

								ZeroMemory(buffer, 10); 
								GetPrivateProfileString(m1.msg, "Y2", NULL, buffer, 5, Path);
								if(isNumeric(buffer))
								{
									newArea->y2=getInteger(buffer,10)*16;
									
									ZeroMemory(buffer, 10); 
									GetPrivateProfileString(m1.msg, "Freq", "-1", buffer, 5, Path);
									newArea->freq=getInteger(buffer,10);

									for(Uint16 j=0;j<5;j++)
									{
										m2="EnterMessage";
										m2+=j;
										ZeroMemory(buffer, 250); 
										GetPrivateProfileString(m1.msg, m2.msg, NULL, buffer, 250, Path);
										if(buffer[0])
										{
											newArea->Enter[j]=buffer;
										}
		
										m2="LeaveMessage";
										m2+=j;
										ZeroMemory(buffer, 250); 
										GetPrivateProfileString(m1.msg, m2.msg, NULL, buffer, 250, Path);
										if(buffer[0])
										{
											newArea->Leave[j]=buffer;
										}
									}

									newArea->AreaNum=i;
									AreaList.append(newArea);
								}
								else
								{goto Delete;}
							}
							else
							{goto Delete;}
						}
						else
						{goto Delete;}
					}
					
					continue;
				Delete:
					delete newArea;
				}
			}
		}
		break;
	case EVENT_Term:
		{
			tell(makeEcho("DLL plugin disconnected."));
		}
		break;
	};
}


//////// DLL export ////////

_declspec(dllexport) void __stdcall talk(BotEvent &event)
{
	botInfo *bot;

	bot = findBot(event.handle);

	if (!bot)
	{
		bot = new botInfo(event.handle);
		botlist.append(bot);
	}

	if (bot) bot->gotEvent(event);
}


//////// Tag emulation ////////

int botInfo::get_tag(Player *p, int index)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
			return tag->data;

		parse = parse->next;
	}

	return 0;
}

void botInfo::set_tag(Player *p, int index, int data)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;

		if (tag->p == p)
		if (tag->index == index)
		{
			tag->data = data;
			return;
		}

		parse = parse->next;
	}

	tag = new PlayerTag;
	tag->p = p;
	tag->index = index;
	tag->data = data;
	taglist.append(tag);
}

void botInfo::killTags(Player *p)
{
	_listnode <PlayerTag> *parse = taglist.head;
	PlayerTag *tag;

	while (parse)
	{
		tag = parse->item;
		parse = parse->next;

		if (tag->p != p) continue;

		taglist.kill(tag);
	}
}

void botInfo::killTags()
{
	taglist.clear();
}


//////// LVZ Object Toggling ////////

void botInfo::clear_objects()
{
	num_objects = 0;
}

void botInfo::object_target(Player *p)
{
	object_dest = p;
}

void botInfo::toggle_objects()
{
	Player *p = object_dest;

	if (!p)	tell(makeToggleObjects(UNASSIGNED, (Uint16 *)object_array, num_objects));
	else	tell(makeToggleObjects(p->ident, (Uint16 *)object_array, num_objects));

	num_objects = 0;
}

void botInfo::queue_enable(int id)
{
	if (num_objects == MAX_OBJECTS)
		toggle_objects();

	object_array[num_objects].id = id;
	object_array[num_objects].disabled = false;
	++num_objects;
}

void botInfo::queue_disable(int id)
{
	if (num_objects == MAX_OBJECTS)
		toggle_objects();

	object_array[num_objects].id = id;
	object_array[num_objects].disabled = true;
	++num_objects;
}


//////// Chatter ////////

void botInfo::sendPrivate(Player *player, BYTE snd, char *msg)
{
	tell(makeSay(MSG_Private, snd, player->ident, msg));
}

void botInfo::sendPrivate(Player *player, char *msg)
{
	tell(makeSay(MSG_Private, 0, player->ident, msg));
}

void botInfo::sendTeam(char *msg)
{
	tell(makeSay(MSG_Team, 0, 0, msg));
}

void botInfo::sendTeam(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Team, snd, 0, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, 0, team, msg));
}

void botInfo::sendTeamPrivate(Uint16 team, BYTE snd, char *msg)
{
	tell(makeSay(MSG_TeamPrivate, snd, team, msg));
}

void botInfo::sendPublic(char *msg)
{
	tell(makeSay(MSG_Public, 0, 0, msg));
}

void botInfo::sendPublic(BYTE snd, char *msg)
{
	tell(makeSay(MSG_Public, snd, 0, msg));
}

void botInfo::sendPublicMacro(char *msg)
{
	tell(makeSay(MSG_PublicMacro, 0, 0, msg));
}

void botInfo::sendPublicMacro(BYTE snd, char *msg)
{
	tell(makeSay(MSG_PublicMacro, snd, 0, msg));
}

void botInfo::sendChannel(char *msg)
{
	tell(makeSay(MSG_Channel, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *msg)
{
	tell(makeSay(MSG_RemotePrivate, 0, 0, msg));
}

void botInfo::sendRemotePrivate(char *name, char *msg)
{
	String s;
	s += ":";
	s += name;
	s += ":";
	s += msg;

	sendRemotePrivate(s);
}
